/*
 *
 * $Id: postgre.c, v 0.9 2001/04/06 23:22:45 kaiser Exp $
 *
 * postgresql backend to ipac
 * Copyright (C) 2001 Al Zaharov
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * The author can be reached via email: kaiser13@mail2000.ru, or by 
 * fido: Al_Zaharov@p88.f58.n5005.z2.fidonet.org
 *
*/

#include <stdio.h>
#include <string.h>
#include <dirent.h>
#include <errno.h>
#include <stdlib.h>
#include <sys/param.h>
#include <unistd.h>
#include <search.h>
#include <assert.h>
#include <netdb.h>
#include <stdarg.h>
#include <fcntl.h>
#include <sys/wait.h>
#include <dlfcn.h>

#include "ipac.h"
#include "libiptc.h"
#include "../../lib/libnet.h"


#ifndef PROC_SYS_MODPROBE
#define PROC_SYS_MODPROBE "/proc/sys/kernel/modprobe"
#endif

#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif

#ifndef IPT_LIB_DIR
#define IPT_LIB_DIR "/lib/iptables"
#endif

#define FMT_NUMERIC	0x0001
#define FMT_NOCOUNTS	0x0002
#define FMT_KILOMEGAGIGA 0x0004
#define FMT_OPTIONS	0x0008
#define FMT_NOTABLE	0x0010
#define FMT_NOTARGET	0x0020
#define FMT_VIA		0x0040
#define FMT_NONEWLINE	0x0080
#define FMT_LINENUMBERS 0x0100

#define FMT_PRINT_RULE (FMT_NOCOUNTS | FMT_OPTIONS | FMT_VIA \
			| FMT_NUMERIC | FMT_NOTABLE)
#define FMT(tab,notab) ((format) & FMT_NOTABLE ? (notab) : (tab))

/* Include file for additions: new matches and targets. */
struct iptables_match
{
	struct iptables_match *next;

	ipt_chainlabel name;

	const char *version;

	/* Size of match data. */
	size_t size;

	/* Size of match data relevent for userspace comparison purposes */
	size_t userspacesize;

	/* Function which prints out usage message. */
	void (*help)(void);

	/* Initialize the match. */
	void (*init)(struct ipt_entry_match *m, unsigned int *nfcache);

	/* Function which parses command options; returns true if it
           ate an option */
	int (*parse)(int c, char **argv, int invert, unsigned int *flags,
		     const struct ipt_entry *entry,
		     unsigned int *nfcache,
		     struct ipt_entry_match **match);

	/* Final check; exit if not ok. */
	void (*final_check)(unsigned int flags);

	/* Prints out the match iff non-NULL: put space at end */
	void (*print)(const struct ipt_ip *ip,
		      const struct ipt_entry_match *match, int numeric);

	/* Saves the match info in parsable form to stdout. */
	void (*save)(const struct ipt_ip *ip,
		     const struct ipt_entry_match *match);

	/* Pointer to list of extra command-line options */
	const struct option *extra_opts;

	/* Ignore these men behind the curtain: */
	unsigned int option_offset;
	struct ipt_entry_match *m;
	unsigned int mflags;
	unsigned int used;
};

struct iptables_target
{
	struct iptables_target *next;

	ipt_chainlabel name;

	const char *version;

	/* Size of target data. */
	size_t size;

	/* Size of target data relevent for userspace comparison purposes */
	size_t userspacesize;

	/* Function which prints out usage message. */
	void (*help)(void);

	/* Initialize the target. */
	void (*init)(struct ipt_entry_target *t, unsigned int *nfcache);

	/* Function which parses command options; returns true if it
           ate an option */
	int (*parse)(int c, char **argv, int invert, unsigned int *flags,
		     const struct ipt_entry *entry,
		     struct ipt_entry_target **target);

	/* Final check; exit if not ok. */
	void (*final_check)(unsigned int flags);

	/* Prints out the target iff non-NULL: put space at end */
	void (*print)(const struct ipt_ip *ip,
		      const struct ipt_entry_target *target, int numeric);

	/* Saves the targinfo in parsable form to stdout. */
	void (*save)(const struct ipt_ip *ip,
		     const struct ipt_entry_target *target);

	/* Pointer to list of extra command-line options */
	struct option *extra_opts;

	/* Ignore these men behind the curtain: */
	unsigned int option_offset;
	struct ipt_entry_target *t;
	unsigned int tflags;
	unsigned int used;
};

enum ipt_tryload {
	DONT_LOAD,
	TRY_LOAD,
	LOAD_MUST_SUCCEED
};


/** structure for run file (rule file) */
struct runfile_line_type {
	char *line;
	struct runfile_line_type *next;
};

static iptc_handle_t handle;

struct iptables_match *iptables_matches = NULL;
struct iptables_target *iptables_targets = NULL;

extern void register_match(struct iptables_match *me);
extern void register_target(struct iptables_target *me);

/* plain file ipac interface entries */
int iptables_ipac_init(int flag);
int iptables_ipac_set(rule_type **firstrule);
int iptables_ipac_read(rule_type **firstrule);
int iptables_ipac_deny(char *login);
int iptables_ipac_accept(char *login);

static const acc_agent_t interface_entry = {
	"iptables",
	iptables_ipac_init,
	iptables_ipac_set,
	iptables_ipac_read,
	iptables_ipac_deny,
	iptables_ipac_accept
};

const acc_agent_t *ipac_ag_interface_iptables() {
	return &interface_entry;
}


//===================================================
static int
tcp_service_to_port(const char *name)
{
	struct servent *service;

	if ((service = getservbyname(name, "tcp")) != NULL)
		return ntohs((unsigned short) service->s_port);

	return -1;
}

static u_int16_t
parse_tcp_port(const char *port)
{
	int portnum;

	if ((portnum = string_to_number(port, 0, 65535)) != -1 ||
	    (portnum = tcp_service_to_port(port)) != -1)
		return (u_int16_t)portnum;
	
	fprintf(stderr, "invalid TCP port/service `%s' specified\n", port);
	exit(1);
}

static int
udp_service_to_port(const char *name)
{
	struct servent *service;

	if ((service = getservbyname(name, "udp")) != NULL)
		return ntohs((unsigned short) service->s_port);

	return -1;
}

static u_int16_t
parse_udp_port(const char *port)
{
	int portnum;

	if ((portnum = string_to_number(port, 0, 65535)) != -1 ||
	    (portnum = udp_service_to_port(port)) != -1)
		return (u_int16_t)portnum;
	
	fprintf(stderr, "invalid UDP port/service `%s' specified\n", port);
	exit(1);
}

// ---------------------------------------------------------------------
// ---------------------------------------------------------------------

static char *
proto_to_name(u_int8_t proto, int nolookup)
{
	unsigned int i;

	if (proto && !nolookup) {
		struct protoent *pent = getprotobynumber(proto);
		if (pent)
			return pent->p_name;
	}

	for (i = 0; i < sizeof(chain_protos)/sizeof(struct pprot); i++)
		if (chain_protos[i].num == proto)
			return chain_protos[i].name;

	return NULL;
}

static char *
addr_to_network(const struct in_addr *addr)
{
	struct netent *net;

	if ((net = getnetbyaddr((long) ntohl(addr->s_addr), AF_INET)) != NULL)
		return (char *) net->n_name;

	return (char *) NULL;
}

char *
addr_to_dotted(const struct in_addr *addrp)
{
	static char buf[20];
	const unsigned char *bytep;

	bytep = (const unsigned char *) &(addrp->s_addr);
	sprintf(buf, "%d.%d.%d.%d", bytep[0], bytep[1], bytep[2], bytep[3]);
	return buf;
}

static char *
addr_to_host(const struct in_addr *addr)
{
	struct hostent *host;

	if ((host = gethostbyaddr((char *) addr,
				  sizeof(struct in_addr), AF_INET)) != NULL)
		return (char *) host->h_name;

	return (char *) NULL;
}

static char *
addr_to_anyname(const struct in_addr *addr)
{
	char *name;

	if ((name = addr_to_host(addr)) != NULL ||
	    (name = addr_to_network(addr)) != NULL)
		return name;

	return addr_to_dotted(addr);
}

static char *
mask_to_dotted(const struct in_addr *mask)
{
	int i;
	static char buf[20];
	u_int32_t maskaddr, bits;

	maskaddr = ntohl(mask->s_addr);

	if (maskaddr == 0xFFFFFFFFL)
		/* we don't want to see "/32" */
		return "";

	i = 32;
	bits = 0xFFFFFFFEL;
	while (--i >= 0 && maskaddr != bits)
		bits <<= 1;
	if (i >= 0)
		sprintf(buf, "/%d", i);
	else
		/* mask was not a decent combination of 1's and 0's */
		sprintf(buf, "/%s", addr_to_dotted(mask));

	return buf;
}

static struct ipt_entry *
generate_entry(const struct ipt_entry *fw,
	       struct iptables_match *matches,
	       struct ipt_entry_target *target)
{
	unsigned int size;
	struct iptables_match *m;
	struct ipt_entry *e;

	size = sizeof(struct ipt_entry);
	for (m = matches; m; m = m->next) {
		if (!m->used)
			continue;

		size += m->m->u.match_size;
	}

	e = xmalloc(size + target->u.target_size);
	*e = *fw;
	e->target_offset = size;
	e->next_offset = size + target->u.target_size;

	size = 0;
	for (m = matches; m; m = m->next) {
		if (!m->used)
			continue;

		memcpy(e->elems + size, m->m, m->m->u.match_size);
		size += m->m->u.match_size;
	}
	memcpy(e->elems + size, target, target->u.target_size);

	return e;
}

static char *get_modprobe(void)
{
	int procfile;
	char *ret;

	procfile = open(PROC_SYS_MODPROBE, O_RDONLY);
	if (procfile < 0)
		return NULL;

	ret = malloc(1024);
	if (ret) {
		switch (read(procfile, ret, 1024)) {
		case -1: goto fail;
		case 1024: goto fail; /* Partial read.  Wierd */
		}
		if (ret[strlen(ret)-1]=='\n') 
			ret[strlen(ret)-1]=0;
		close(procfile);
		return ret;
	}
 fail:
	free(ret);
	close(procfile);
	return NULL;
}

struct iptables_target *
find_target(const char *name, enum ipt_tryload tryload)
{
	struct iptables_target *ptr;

//	printf("Yes! find_target is called! and name is: %s\n", name);

	/* Standard target? */
	if (strcmp(name, "") == 0
	    || strcmp(name, IPTC_LABEL_ACCEPT) == 0
	    || strcmp(name, IPTC_LABEL_DROP) == 0
	    || strcmp(name, IPTC_LABEL_QUEUE) == 0
	    || strcmp(name, IPTC_LABEL_RETURN) == 0)
		name = "standard";


	for (ptr = iptables_targets; ptr; ptr = ptr->next) {
		if (strcmp(name, ptr->name) == 0)
		{
			break;
		}
	}

	if (!ptr && tryload != DONT_LOAD) {
		char path[sizeof(IPT_LIB_DIR) + sizeof("/libipt_.so")
			 + strlen(name)];
		sprintf(path, IPT_LIB_DIR "/libipt_%s.so", name);
		if (dlopen(path, RTLD_NOW)) {
			/* Found library.  If it didn't register itself,
			   maybe they specified match as a target. */
			ptr = find_target(name, DONT_LOAD);
			if (!ptr) {
				fprintf(stderr, "Couldn't load target `%s'\n",
					   name);
				exit(1);
			}
		} else if (tryload == LOAD_MUST_SUCCEED) {
			fprintf(stderr, "Couldn't load target `%s':%s\n",
				   name, dlerror());
			exit(1);
			}
	}

	if (ptr)
		ptr->used = 1;

	return ptr;
}

static int iptables_insmod(const char *modname, const char *modprobe)
{
	char *buf = NULL;
	char *argv[3];

//	 If they don't explicitly set it, read out of kernel 
	if (!modprobe) {
		buf = get_modprobe();
		if (!buf)
			return -1;
		modprobe = buf;
	}

	switch (fork()) {
	case 0:
		argv[0] = (char *)modprobe;
		argv[1] = (char *)modname;
		argv[2] = NULL;
		execv(argv[0], argv);

//		 not usually reached 
		exit(0);
	case -1:
		return -1;

	default: // parent 
		wait(NULL);
	}

	free(buf);
	return 0;
}

void
register_target(struct iptables_target *me)
{
//	if (strcmp(me->version, program_version) != 0) {
//		fprintf(stderr, "%s: target `%s' v%s (I'm v%s).\n",
//			program_name, me->name, me->version, program_version);
//		exit(1);
//	}
//	printf("Yes! register_target were called!\n");

	if (find_target(me->name, DONT_LOAD)) {
		fprintf(stderr, "%s: target `%s' already registered.\n",
			"fddfewev", me->name);
		exit(1);
	}

	if (me->size != IPT_ALIGN(me->size)) {
		fprintf(stderr, "%s: target `%s' has invalid size %u.\n",
			"fddfgdsse", me->name, me->size);
		exit(1);
	}

//	Prepend to list.
	me->next = iptables_targets;
	iptables_targets = me;
	me->t = NULL;
	me->tflags = 0;
}

unsigned char * make_delete_mask(struct ipt_entry *fw)
{
	/* Establish mask for comparison */
	unsigned int size;
	struct iptables_match *m;
	unsigned char *mask, *mptr;

	size = sizeof(struct ipt_entry);
	for (m = iptables_matches; m; m = m->next) {
		if (!m->used)
			continue;

		size += IPT_ALIGN(sizeof(struct ipt_entry_match)) + m->size;
	}

	mask = xcalloc(1, size
			 + IPT_ALIGN(sizeof(struct ipt_entry_target))
			 + iptables_targets->size);

	memset(mask, 0xFF, sizeof(struct ipt_entry));
	mptr = mask + sizeof(struct ipt_entry);

	for (m = iptables_matches; m; m = m->next) {
		if (!m->used)
			continue;

		memset(mptr, 0xFF,
		       IPT_ALIGN(sizeof(struct ipt_entry_match))
		       + m->userspacesize);
		mptr += IPT_ALIGN(sizeof(struct ipt_entry_match)) + m->size;
	}

	memset(mptr, 0xFF,
	       IPT_ALIGN(sizeof(struct ipt_entry_target))
	       + iptables_targets->userspacesize);

	return mask;
}

struct iptables_match *
find_match(const char *name, enum ipt_tryload tryload)
{
	struct iptables_match *ptr;
    
//	printf("Yes! find_match is called! and name is: %s\n", name);

	for (ptr = iptables_matches; ptr; ptr = ptr->next) {
		if (strcmp(name, ptr->name) == 0)
			break;
	}

	if (!ptr && tryload != DONT_LOAD) {
		char path[sizeof(IPT_LIB_DIR) + sizeof("/libipt_.so")
			 + strlen(name)];
		sprintf(path, IPT_LIB_DIR "/libipt_%s.so", name);
		if (dlopen(path, RTLD_NOW)) {
			/* Found library.  If it didn't register itself,
			   maybe they specified target as match. */
			ptr = find_match(name, DONT_LOAD);

			if (!ptr) {
				fprintf(stderr, "Couldn't load match `%s'\n",
					   name);
				exit(1);
			}
		} else if (tryload == LOAD_MUST_SUCCEED) {
			fprintf(stderr, "Couldn't load match `%s':%s\n",
				   name, dlerror());
			exit(1);
		}
	}

	if (ptr)
		ptr->used = 1;

	return ptr;
}

void
register_match(struct iptables_match *me)
{
	struct iptables_match **i;

//	printf("Yes! register_match were called!\n");
/*
	if (strcmp(me->version, program_version) != 0) {
		fprintf(stderr, "%s: match `%s' v%s (I'm v%s).\n",
			program_name, me->name, me->version, program_version);
		exit(1);
	}
*/
	if (find_match(me->name, DONT_LOAD)) {
		fprintf(stderr, "%s: match `%s' already registered.\n",
			"fetchipac??", me->name);
		exit(1);
	}

	if (me->size != IPT_ALIGN(me->size)) {
		fprintf(stderr, "%s: match `%s' has invalid size %u.\n",
			"fetchipac??", me->name, me->size);
		exit(1);
	}

	/* Append to list. */
	for (i = &iptables_matches; *i; i = &(*i)->next);
	me->next = NULL;
	*i = me;

	me->m = NULL;
	me->mflags = 0;
}


static struct iptables_match *
find_proto(const char *pname, enum ipt_tryload tryload, int nolookup)
{
	int proto;

	proto = string_to_number(pname, 0, 255);
	if (proto != -1)
		return find_match(proto_to_name(proto, nolookup), tryload);

	return find_match(pname, tryload);
}

static void
print_num(u_int64_t number, unsigned int format)
{
	if (format & FMT_KILOMEGAGIGA) {
		if (number > 99999) {
			number = (number + 500) / 1000;
			if (number > 9999) {
				number = (number + 500) / 1000;
				if (number > 9999) {
					number = (number + 500) / 1000;
					printf(FMT("%4lluG ","%lluG "),number);
				}
				else printf(FMT("%4lluM ","%lluM "), number);
			} else
				printf(FMT("%4lluK ","%lluK "), number);
		} else
			printf(FMT("%5llu ","%llu "), number);
	} else
		printf(FMT("%8llu ","%llu "), number);
}

static int
print_match(const struct ipt_entry_match *m,
	    const struct ipt_ip *ip,
	    int numeric)
{
	struct iptables_match *match = find_match(m->u.user.name, TRY_LOAD);

	if (match) {
		if (match->print)
			match->print(ip, m, numeric);
		else
			printf("%s ", match->name);
	} else {
		if (m->u.user.name[0])
			printf("UNKNOWN match `%s' ", m->u.user.name);
	}
	/* Don't stop iterating. */
	return 0;
}

/* e is called `fw' here for hysterical raisins */
static void
print_firewall(const struct ipt_entry *fw,
	       const char *targname,
	       unsigned int num,
	       unsigned int format,
	       const iptc_handle_t handle)
{
	struct iptables_target *target = NULL;
	const struct ipt_entry_target *t;
	u_int8_t flags;
	char buf[BUFSIZ];

	/* User creates a chain called "REJECT": this overrides the
	   `REJECT' target module.  Keep feeding them rope until the
	   revolution... Bwahahahahah */
	if (!iptc_is_chain(targname, handle))
		target = find_target(targname, TRY_LOAD);
	else
		target = find_target(IPT_STANDARD_TARGET, LOAD_MUST_SUCCEED);

	t = ipt_get_target((struct ipt_entry *)fw);
	flags = fw->ip.flags;

	if (format & FMT_LINENUMBERS)
		printf(FMT("%-4u ", "%u "), num+1);

	if (!(format & FMT_NOCOUNTS)) {
		print_num(fw->counters.pcnt, format);
		print_num(fw->counters.bcnt, format);
	}

	if (!(format & FMT_NOTARGET))
		printf(FMT("%-9s ", "%s "), targname);

	fputc(fw->ip.invflags & IPT_INV_PROTO ? '!' : ' ', stdout);
	{
		char *pname = proto_to_name(fw->ip.proto, format&FMT_NUMERIC);
		if (pname)
			printf(FMT("%-5s", "%s "), pname);
		else
			printf(FMT("%-5hu", "%hu "), fw->ip.proto);
	}

	if (format & FMT_OPTIONS) {
		if (format & FMT_NOTABLE)
			fputs("opt ", stdout);
		fputc(fw->ip.invflags & IPT_INV_FRAG ? '!' : '-', stdout);
		fputc(flags & IPT_F_FRAG ? 'f' : '-', stdout);
		fputc(' ', stdout);
	}

	if (format & FMT_VIA) {
		char iface[IFNAMSIZ+2];

		if (fw->ip.invflags & IPT_INV_VIA_IN) {
			iface[0] = '!';
			iface[1] = '\0';
		}
		else iface[0] = '\0';

		if (fw->ip.iniface[0] != '\0') {
			strcat(iface, fw->ip.iniface);
			/* If it doesn't compare the nul-term, it's a
			   wildcard. */
			if (fw->ip.iniface_mask[strlen(fw->ip.iniface)] == 0)
				strcat(iface, "+");
		}
		else if (format & FMT_NUMERIC) strcat(iface, "*");
		else strcat(iface, "any");
		printf(FMT(" %-6s ","in %s "), iface);

		if (fw->ip.invflags & IPT_INV_VIA_OUT) {
			iface[0] = '!';
			iface[1] = '\0';
		}
		else iface[0] = '\0';

		if (fw->ip.outiface[0] != '\0') {
			strcat(iface, fw->ip.outiface);
			/* If it doesn't compare the nul-term, it's a
			   wildcard. */
			if (fw->ip.outiface_mask[strlen(fw->ip.outiface)] == 0)
				strcat(iface, "+");
		}
		else if (format & FMT_NUMERIC) strcat(iface, "*");
		else strcat(iface, "any");
		printf(FMT("%-6s ","out %s "), iface);
	}

	fputc(fw->ip.invflags & IPT_INV_SRCIP ? '!' : ' ', stdout);
	if (fw->ip.smsk.s_addr == 0L && !(format & FMT_NUMERIC))
		printf(FMT("%-19s ","%s "), "anywhere");
	else {
		if (format & FMT_NUMERIC)
			sprintf(buf, "%s", addr_to_dotted(&(fw->ip.src)));
		else
			sprintf(buf, "%s", addr_to_anyname(&(fw->ip.src)));
		strcat(buf, mask_to_dotted(&(fw->ip.smsk)));
		printf(FMT("%-19s ","%s "), buf);
	}

	fputc(fw->ip.invflags & IPT_INV_DSTIP ? '!' : ' ', stdout);
	if (fw->ip.dmsk.s_addr == 0L && !(format & FMT_NUMERIC))
		printf(FMT("%-19s","-> %s"), "anywhere");
	else {
		if (format & FMT_NUMERIC)
			sprintf(buf, "%s", addr_to_dotted(&(fw->ip.dst)));
		else
			sprintf(buf, "%s", addr_to_anyname(&(fw->ip.dst)));
		strcat(buf, mask_to_dotted(&(fw->ip.dmsk)));
		printf(FMT("%-19s","-> %s"), buf);
	}

	if (format & FMT_NOTABLE)
		fputs("  ", stdout);

	IPT_MATCH_ITERATE(fw, print_match, &fw->ip, format & FMT_NUMERIC);

	if (target) {
		if (target->print)
			/* Print the target information. */
			target->print(&fw->ip, t, format & FMT_NUMERIC);
	} else if (t->u.target_size != sizeof(*t))
		printf("[%u bytes of unknown target data] ",
		       t->u.target_size - sizeof(*t));

	if (!(format & FMT_NONEWLINE))
		fputc('\n', stdout);
}

static void
print_firewall_line(const struct ipt_entry *fw,
		    const iptc_handle_t h)
{
	struct ipt_entry_target *t;

	t = ipt_get_target((struct ipt_entry *)fw);
	print_firewall(fw, t->u.user.name, 0, FMT_PRINT_RULE, h);
}


void
exit_error(enum exittype status, char *msg, ...)
{
	va_list args;

	va_start(args, msg);
//	fprintf(stderr, "%s v%s: ", program_name, program_version);
	vfprintf(stderr, msg, args);
	va_end(args);
	fprintf(stderr, "\n");
//	if (status == PARAMETER_PROBLEM)
//		exit_tryhelp(status);
	if (status == VERSION_PROBLEM)
		fprintf(stderr,
			"Perhaps iptables or your kernel needs to be upgraded.\n");
	exit(status);
}

int
check_inverse(const char option[], int *invert)
{
	if (option && strcmp(option, "!") == 0) {
		if (*invert)
			exit_error(PARAMETER_PROBLEM,
				   "Multiple `!' flags not allowed");

		*invert = TRUE;
		return TRUE;
	}
	return FALSE;
}

// ---------------------------------------------------------------------


/** delete run file (rule file) from memory, freeing dynamically
 *  allocated memory
 */
static void 
destroy_runfile_lines(struct runfile_line_type *lines)
{
	struct runfile_line_type *next;
	while(lines != NULL)
	{
		if (lines->line != NULL)
			free(lines->line);
		next = lines->next;
		free(lines);
		lines = next;
	}
}

//------------------------------------------------------------------
/** read run file (rule file) and store it in memory, using a singly 
 *  linked list of instances of struct runfile_line_type
 *  return the list read or NULL in case of error
 */
static struct runfile_line_type 
*read_runfile()
{
	FILE *frunfile;
	char runfile_line[MAX_RULE_NAME_LENGTH + 50], *cp; 
	struct runfile_line_type *result, *lastline, *cur;
	
	int tmp=0;

	frunfile = fopen(RUNFILE, "r");
	if (frunfile == NULL)
	{
		fprintf(stderr, "%s: cant open run file \"%s\": %s "
				"(ipacset not run?)\n",
			me, RUNFILE, sys_errlist[errno]);
		return NULL;
	}

	result = NULL;
	lastline = NULL;
	while(fgets(runfile_line, MAX_RULE_NAME_LENGTH+50, frunfile)
			!= NULL)
	{
	    tmp++;
		cp = strchr(runfile_line, '\n');
		if (cp)
			*cp = '\0';
		if (*runfile_line == '#')
			continue;

		cur = (struct runfile_line_type *)
				xmalloc(sizeof(struct runfile_line_type));
		cur->line = xstrdup(runfile_line);
		cur->next = NULL;
		if (result == NULL)
			result = cur;
		else 
			lastline->next = cur;
		lastline = cur;
	}
	if (!feof(frunfile))
	{
		fprintf(stderr, "%s: reading \"%s\": %s\n",
			me, RUNFILE, sys_errlist[errno]);
		fclose(frunfile);
		destroy_runfile_lines(result);
		result = NULL;
	}
	fclose(frunfile);
	return result;
}

/** read kernel accounting data in iptables system
 * read from stream f
 * create records with data (packet and data counters) and
 * rule names and store them into instances of rule_type (declared
 * in ipac.h) using rule names from runfile
 * if a rule name is equal to a previously used rule name, add the
 * counters and re-use the old record
 * complain if something is wrong with the data.
 * return 0 for success, 1 otherwise
 */
int read_iptables(struct runfile_line_type *runfile, rule_type **firstrule)
{
	char *cp="\0";
	rule_type *rule, *lastrule;
	struct runfile_line_type *nextline;
	void *node;
	void *ruletree = NULL;
	char some[10];
	int inb=0, outb=0;
	struct ipt_counters *counters = NULL;
	const char *modprobe = NULL;


	if (!handle)
		handle = iptc_init("filter");

	if (!handle) {
//		try to insmod the module if iptc_init failed
		iptables_insmod("ip_tables", modprobe);
		handle = iptc_init("filter");
	}

	if (!handle) {
		perror("can't initialize iptables table `filter'\n");
		return (1);
	}

	/* create the rule_type records in correct order as from 
	 * run file.
	 */
	lastrule = *firstrule = NULL;

	for (nextline=runfile; nextline!=NULL; nextline=nextline->next)
	{
		cp = strchr(nextline->line, ' ');
		if (cp == 0)
			continue;	/* bad entry */
		rule = new_rule();

		strncpy(rule->name, cp+1, MAX_RULE_NAME_LENGTH);
		strncpy(some, nextline->line, cp-nextline->line);
		some[cp-nextline->line]='\0';
//		printf("some: %s, CH_INNAME: %s, CH_OUTNAME: %s.\n", some,
//						    CH_INNAME, CH_OUTNAME);
		if (!strcmp(some, CH_INNAME)) {
			inb++;
			counters = iptc_read_counter(CH_INNAME, inb-1, &handle);
			if (!iptc_zero_counter(CH_INNAME, inb-1, &handle))
			    fprintf(stderr, "iptables: %s\n", iptc_strerror(errno));
			}
		if (!strcmp(some, CH_OUTNAME)) {
			outb++;
			counters = iptc_read_counter(CH_OUTNAME, outb-1, &handle);
			if (!iptc_zero_counter(CH_OUTNAME, outb-1, &handle))
			    fprintf(stderr, "iptables: %s\n", iptc_strerror(errno));
			}
		/* use a binary tree to find rules with same name */
		node = tsearch(rule, &ruletree, rule_compare);
		if (*(rule_type **)node != rule)
		{
			free(rule);
			rule=*(rule_type **)node;
		} else
		{
			if (lastrule != NULL)
				lastrule->next = rule;
			lastrule = rule;
			if (*firstrule == NULL)
				*firstrule = rule;
		}
		rule->pkts += counters->pcnt;
		rule->bytes += counters->bcnt;
	}
	iptc_commit(&handle);
	free_tree(&ruletree);
	return 0;
}
/*
 * Try to insert rule into kernel return 0 in case all right, 1 otherwice
 */
int
append_rule (char *chain, char *saddr, char *sport, char *daddr, char *dport,
		char *proto, char *targ)
{
	struct ipt_entry fw, *e = NULL;
	unsigned int naddrs = 0;
	struct in_addr *addrs = NULL;
	struct iptables_match *m;
	struct iptables_target *target = NULL;
	struct iptables_target *t;
	size_t size;

	memset(&fw, 0, sizeof(fw));

	if (!memcmp(proto, "all", 3))
		strcpy(proto, "");

	for (m = iptables_matches; m; m = m->next) {
		m->mflags = 0;
		m->used = 0;
	}

	for (t = iptables_targets; t; t = t->next) {
		t->tflags = 0;
		t->used = 0;
	}
	
	if (!iptc_is_chain(chain, handle)) {
		fprintf(stderr, "%s is not a chain\n", chain);
		return (1);
	}
	
	if (saddr[0]=='\0') 
		saddr="0.0.0.0/0";
	if (daddr[0]=='\0') 
		daddr="0.0.0.0/0";
	parse_hostnetworkmask(saddr, &addrs, &(fw.ip.smsk), &naddrs);
	if (naddrs>1) {
		fprintf(stderr, "Bad rule: more than 1 source address\n");
		return (1);
	}
	fw.ip.src.s_addr = addrs[0].s_addr;
	fw.nfcache |= NFC_IP_SRC;
	parse_hostnetworkmask(daddr, &addrs, &(fw.ip.dmsk), &naddrs);
	if (naddrs>1) {
		fprintf(stderr, "Incorrect rule: more than 1 destination address\n");
		return (1);
	}
	fw.ip.dst.s_addr = addrs[0].s_addr;
	fw.nfcache |= NFC_IP_DST;

	if ((sport[0]!='\0' || dport[0]!='\0') && proto[0]=='\0') {
		fprintf(stderr, "Incorrect rule: source or destination port spe"
		    "cified while protocol is not\n");
		return (1);
	}
	
	if (proto[0] != '\0') {
		fw.ip.proto = parse_protocol(proto);
		fw.nfcache |= NFC_IP_PROTO;
	}
	
	/* Loading target /if any/ */
	target = find_target(IPT_STANDARD_TARGET, LOAD_MUST_SUCCEED);
	size = sizeof(struct ipt_entry_target) + target->size;
	target->t = xcalloc(1, size);
	target->t->u.target_size = size;
	strcpy(target->t->u.user.name, targ);
	target->init(target->t, &fw.nfcache);

	if (proto[0] != '\0') 
	{
		m = find_proto(proto, LOAD_MUST_SUCCEED, 0);
		size = IPT_ALIGN(sizeof(struct ipt_entry_match)) + m->size;
		m->m = xcalloc(size, 1);
		m->m->u.match_size = size;
		strcpy(m->m->u.user.name, m->name);
		m->init(m->m, &fw.nfcache);
		if (sport[0]!='\0' || dport[0]!='\0') 
		{
			if (!strcmp(proto, "tcp")) {
				struct ipt_tcp *tcpinfo = 
						(struct ipt_tcp *)m->m->data;
				if (sport[0]!='\0')
					tcpinfo->spts[0] = tcpinfo->spts[1] = 
							parse_tcp_port(sport);
				if (dport[0]!='\0')
					tcpinfo->dpts[0] = tcpinfo->dpts[1] = 
							parse_tcp_port(dport);
			}
			if (!strcmp(proto, "udp")) {
				struct ipt_udp *udpinfo = 
						(struct ipt_udp *)m->m->data;
				if (sport[0]!='\0')
					udpinfo->spts[0] = udpinfo->spts[1] = 
							parse_udp_port(sport);
				if (dport[0]!='\0')
					udpinfo->dpts[0] = udpinfo->dpts[1] = 
							parse_udp_port(dport);
			}
		}
	}
	target->final_check(target->tflags);
	e = generate_entry(&fw, iptables_matches, target->t);

	if (!handle) if (!(handle = iptc_init("filter"))) {
			    fprintf(stderr, "iptables: %s\n", iptc_strerror(errno));
			    return (1);
			}
	if (verbose) {
		printf("Appending rule\n");
		print_firewall_line(e, handle);
	}
	if (!iptc_append_entry(chain, e, &handle)) {
		fprintf(stderr, "iptables: %s\n", iptc_strerror(errno));
		return (1);
	}
	free(e);
	return 0;
}

int
delete_rule (char *chain, char *saddr, char *sport, char *daddr, char *dport,
		char *proto, char *targ)
{
	struct ipt_entry fw, *e = NULL;
	unsigned int naddrs = 0;
	struct in_addr *addrs = NULL;
	struct iptables_match *m;
	struct iptables_target *target = NULL;
	struct iptables_target *t;
	unsigned char *mask = NULL;
	size_t size;

	memset(&fw, 0, sizeof(fw));

	if (!memcmp(proto, "all", 3))
		strcpy(proto, "");

	for (m = iptables_matches; m; m = m->next) {
		m->mflags = 0;
		m->used = 0;
	}

	for (t = iptables_targets; t; t = t->next) {
		t->tflags = 0;
		t->used = 0;
	}
	
	if (!iptc_is_chain(chain, handle)) {
		fprintf(stderr, "%s is not a chain\n", chain);
		return (1);
	}
	
	if (saddr[0]=='\0') 
		saddr="0.0.0.0/0";
	if (daddr[0]=='\0') 
		daddr="0.0.0.0/0";
	parse_hostnetworkmask(saddr, &addrs, &(fw.ip.smsk), &naddrs);
	if (naddrs>1) {
		fprintf(stderr, "Bad rule: more than 1 source address\n");
		return (1);
	}
	fw.ip.src.s_addr = addrs[0].s_addr;
	fw.nfcache |= NFC_IP_SRC;
	parse_hostnetworkmask(daddr, &addrs, &(fw.ip.dmsk), &naddrs);
	if (naddrs>1) {
		fprintf(stderr, "Incorrect rule: more than 1 destination address\n");
		return (1);
	}
	fw.ip.dst.s_addr = addrs[0].s_addr;
	fw.nfcache |= NFC_IP_DST;

	if ((sport[0]!='\0' || dport[0]!='\0') && proto[0]=='\0') {
		fprintf(stderr, "Incorrect rule: source or destination port spe"
		    "cified while protocol is not\n");
		return (1);
	}
	
	if (proto[0] != '\0') {
		fw.ip.proto = parse_protocol(proto);
		fw.nfcache |= NFC_IP_PROTO;
	}
	
	/* Loading target /if any/ */
	target = find_target(IPT_STANDARD_TARGET, LOAD_MUST_SUCCEED);
	size = sizeof(struct ipt_entry_target) + target->size;
	target->t = xcalloc(1, size);
	target->t->u.target_size = size;
	strcpy(target->t->u.user.name, targ);
	target->init(target->t, &fw.nfcache);

	if (proto[0] != '\0') 
	{
		m = find_proto(proto, LOAD_MUST_SUCCEED, 0);
		size = IPT_ALIGN(sizeof(struct ipt_entry_match)) + m->size;
		m->m = xcalloc(size, 1);
		m->m->u.match_size = size;
		strcpy(m->m->u.user.name, m->name);
		m->init(m->m, &fw.nfcache);
		if (sport[0]!='\0' || dport[0]!='\0') 
		{
			if (!strcmp(proto, "tcp")) {
				struct ipt_tcp *tcpinfo = 
						(struct ipt_tcp *)m->m->data;
				if (sport[0]!='\0')
					tcpinfo->spts[0] = tcpinfo->spts[1] = 
							parse_tcp_port(sport);
				if (dport[0]!='\0')
					tcpinfo->dpts[0] = tcpinfo->dpts[1] = 
							parse_tcp_port(dport);
			}
			if (!strcmp(proto, "udp")) {
				struct ipt_udp *udpinfo = 
						(struct ipt_udp *)m->m->data;
				if (sport[0]!='\0')
					udpinfo->spts[0] = udpinfo->spts[1] = 
							parse_udp_port(sport);
				if (dport[0]!='\0')
					udpinfo->dpts[0] = udpinfo->dpts[1] = 
							parse_udp_port(dport);
			}
		}
	}
	target->final_check(target->tflags);
	e = generate_entry(&fw, iptables_matches, target->t);

	if (!handle) if (!(handle = iptc_init("filter"))) {
			    fprintf(stderr, "iptables: %s\n", iptc_strerror(errno));
			    return (1);
			}
	if (verbose) {
		printf("Deleting rule\n");
		print_firewall_line(e, handle);
	}

	mask = make_delete_mask(e);
	if (!iptc_delete_entry(chain, e, mask, &handle)) {
		fprintf(stderr, "iptables: %s\n", iptc_strerror(errno));
		return (1);
	}
	free(e);
	return 0;
}

/** Setup chains if they doesn't exist 
 *
 */
int 
setup_tables(void)
{
	if (!iptc_is_chain(CH_INNAME, handle))
		if (!iptc_create_chain(CH_INNAME, &handle)) {
			fprintf(stderr, "iptables: %s\n", iptc_strerror(errno));
			return (1);
		}
	iptc_create_chain(CH_OUTNAME, &handle);
	iptc_create_chain(CH_IONAME, &handle);

	if (!is_classic) {
		iptc_create_chain(CH_CTRL_IN, &handle);
		iptc_create_chain(CH_CTRL_OUT, &handle);
	}
	return 0;
};


int
flush_acc_tables(void)
{
	if (iptc_is_chain(CH_INNAME, handle))
		if (!iptc_flush_entries(CH_INNAME, &handle)) {
			fprintf(stderr, "iptables: %s\n", iptc_strerror(errno));
			return (1);
		}
	iptc_flush_entries(CH_OUTNAME, &handle);
	iptc_flush_entries(CH_IONAME, &handle);
	if (!is_classic) {
		iptc_flush_entries(CH_CTRL_IN, &handle);
		iptc_flush_entries(CH_CTRL_OUT, &handle);
	}
	
	return 0;
}


int iptables_ipac_init(int flag)
{
	const char *modprobe = NULL;

	if (!handle)
		handle = iptc_init("filter");

	if (!handle) {
//		try to insmod the module if iptc_init failed
		iptables_insmod("ip_tables", modprobe);
		handle = iptc_init("filter");
	}

	if (!handle) {
		fprintf(stderr, "ipac-ng: can't initialize iptables table `filter'\n"
		"\tis \"Network packet filtering (replaces ipchains)\"\n"
		"\tin Networking options of your kernel enabled?\n"
		"\tif so then you *have* to enable \"IP tables support\" in\n"
		"\t\"IP: Netfilter Configuration\" *and* \"Packet filtering\"\n"
		"\tsome lines below in your kernel configuration.\n"
		"\tIf you failed to enable these features don't send bug reports\n"
		"\tto me! wbw, kaiser.\n"
		"\tiptables reported that %s\n", iptc_strerror(errno));
		return (1);
	}
	return 0;
};

/** 
 Setup all possible rules in iptables 
 */
int iptables_ipac_set(rule_type **firstrule)
{
	raw_rule_type *d;
	char targ[8];
	char chain[9];
	FILE *frunfile;
	unsigned int ref;


	if (verbose)
		fprintf(stderr, "Flushing accounting chains..\n");
	flush_acc_tables();
	if (verbose)
		fprintf(stderr, "Setting up acc chains..\n");
	setup_tables();
	storage_method->get_raw_list("iptables", "", &d);
	
	frunfile = fopen(RUNFILE, "w");	//needs to make some error handling later

	while(d)
	{
		if (memcmp(d->dest, "in", 2)) 
			strcpy(chain, CH_OUTNAME);
		else
			strcpy(chain, CH_INNAME);
		if (is_classic)
			strcpy(targ, "");
		else {
			if (strcmp(d->name, "admin"))
				strcpy(targ, "");
			else
				strcpy(targ, "REJECT");
			if (!strcmp(d->dest, "in"))
				append_rule(CH_CTRL_IN, d->snet, d->sport,
					d->dnet, d->dport, d->protocol, targ);
			else
				append_rule(CH_CTRL_OUT, d->snet, d->sport,
					d->dnet, d->dport, d->protocol, targ);
			strcpy(targ, "RETURN");
		}
		append_rule(chain, d->snet, d->sport, d->dnet, d->dport,
							    d->protocol, targ);
		fprintf(frunfile,"%s %s\n", chain, d->name);
		d=d->next;
	}
	fclose(frunfile);
	if (!is_classic) {
		iptc_get_references(&ref, CH_CTRL_IN, &handle);
		if (ref!=0)
			delete_rule("INPUT", "", "", "", "", "", CH_CTRL_IN);
		iptc_get_references(&ref, CH_CTRL_OUT, &handle);
		if (ref!=0)
			delete_rule("OUTPUT", "", "", "", "", "", CH_CTRL_OUT);
		append_rule("INPUT", "", "", "", "", "", CH_CTRL_IN);
		append_rule("OUTPUT", "", "", "", "", "", CH_CTRL_OUT);
	}
	iptc_get_references(&ref, CH_INNAME, &handle);
	if (ref!=0) {
		delete_rule("INPUT", "0/0", "", "0/0", "", "", CH_INNAME);
		delete_rule("FORWARD", "0/0", "", "0/0", "", "", CH_INNAME);
	}
	iptc_get_references(&ref, CH_OUTNAME, &handle);
	if (ref!=0) {
		delete_rule("OUTPUT", "0/0", "", "0/0", "", "", CH_OUTNAME);
		delete_rule("FORWARD", "0/0", "", "0/0", "", "", CH_OUTNAME);
	}
	iptc_get_references(&ref, CH_IONAME, &handle);
	if (ref!=0) {
		delete_rule("INPUT", "0/0", "", "0/0", "", "", CH_IONAME);
		delete_rule("OUTPUT", "0/0", "", "0/0", "", "", CH_IONAME);
		delete_rule("FORWARD", "0/0", "", "0/0", "", "", CH_IONAME);
	}
	append_rule("INPUT", "0/0", "", "0/0", "", "", CH_INNAME);
	append_rule("INPUT", "0/0", "", "0/0", "", "", CH_IONAME);
	append_rule("OUTPUT", "0/0", "", "0/0", "", "", CH_OUTNAME);
	append_rule("OUTPUT", "0/0", "", "0/0", "", "", CH_IONAME);
	append_rule("FORWARD", "0/0", "", "0/0", "", "", CH_INNAME);
	append_rule("FORWARD", "0/0", "", "0/0", "", "", CH_OUTNAME);
	append_rule("FORWARD", "0/0", "", "0/0", "", "", CH_IONAME);
	iptc_commit(&handle);
	return 0;
}

int iptables_ipac_read(rule_type **firstrule)
{
	struct runfile_line_type *runfile;

	runfile = read_runfile();
	if (runfile == NULL)
		return 1;

	return read_iptables(runfile, firstrule);
}

int 
iptables_ipac_accept(char *login)
{
/*
	static struct ip_fwuser fw;
	raw_rule_type *dat;
	unsigned int nsaddrs = 0;
	struct in_addr *saddrs = NULL, *daddrs = NULL;
	
	storage_method->get_raw_list("iptables", login, &dat);
	while(dat)
	{
		parse_hostnetworkmask(dat->snet, 
				&saddrs, &(fw.ipfw.fw_smsk), &nsaddrs);	
		parse_hostnetworkmask(dat->dnet, &daddrs,
				&(fw.ipfw.fw_dmsk), &nsaddrs);
		if (!dat->protocol[0]=='\0')
			fw.ipfw.fw_proto = parse_protocol(dat->protocol);
		if (fw.ipfw.fw_proto != IPPROTO_ICMP)
		{
			parse_ports(dat->sport, fw.ipfw.fw_spts, 
							    fw.ipfw.fw_proto);
			parse_ports(dat->dport, fw.ipfw.fw_dpts, 
							    fw.ipfw.fw_proto);
		}
		strcpy(fw.label, "");
		fw.ipfw.fw_src.s_addr = saddrs[0].s_addr;
		fw.ipfw.fw_dst.s_addr = daddrs[0].s_addr;
*/		
/*		
		if (!memcmp(dat->dest, "out", 3))
		{
			strcpy(fw.label, "REJECT");
			iptc_delete_entry(CH_CTRL_OUT, &fw);
			strcpy(fw.label, "");
			iptc_delete_entry(CH_CTRL_OUT, &fw);
			iptc_append_entry(CH_CTRL_OUT, &fw);
		}
		else
		{
			strcpy(fw.label, "REJECT");
			iptc_delete_entry(CH_CTRL_IN, &fw);
			strcpy(fw.label, "");
			iptc_delete_entry(CH_CTRL_IN, &fw);
			iptc_append_entry(CH_CTRL_IN, &fw);
		}
*/		
//		dat=dat->next;
//	}
	return 0;
}

int 
iptables_ipac_deny(char *login)
{
//	static struct ip_fwuser fw;
//	raw_rule_type *dat;
//	unsigned int nsaddrs = 0;
//	struct in_addr *saddrs = NULL, *daddrs = NULL;
/*	
	storage_method->get_raw_list("iptables", login, &dat);
	while(dat)
	{
		parse_hostnetworkmask(dat->snet, 
				&saddrs, &(fw.ipfw.fw_smsk), &nsaddrs);	
		parse_hostnetworkmask(dat->dnet, &daddrs,
				&(fw.ipfw.fw_dmsk), &nsaddrs);
		if (!dat->protocol[0]=='\0')
			fw.ipfw.fw_proto = parse_protocol(dat->protocol);
		if (fw.ipfw.fw_proto != IPPROTO_ICMP)
		{
			parse_ports(dat->sport, fw.ipfw.fw_spts, 
							    fw.ipfw.fw_proto);
			parse_ports(dat->dport, fw.ipfw.fw_dpts, 
							    fw.ipfw.fw_proto);
		}
		strcpy(fw.label, "");
		fw.ipfw.fw_src.s_addr = saddrs[0].s_addr;
		fw.ipfw.fw_dst.s_addr = daddrs[0].s_addr;
*/		
/*		
		if (!memcmp(dat->dest, "out", 3))
		{
			strcpy(fw.label, "");
			iptc_delete_entry(CH_CTRL_OUT, &fw);
			strcpy(fw.label, "REJECT");
			iptc_delete_entry(CH_CTRL_OUT, &fw);
			iptc_append_entry(CH_CTRL_OUT, &fw);
		}
		else
		{
			strcpy(fw.label, "");
			iptc_delete_entry(CH_CTRL_IN, &fw);
			strcpy(fw.label, "REJECT");
			iptc_delete_entry(CH_CTRL_IN, &fw);
			iptc_append_entry(CH_CTRL_IN, &fw);
		}
*/
//		dat=dat->next;
//	}
	return 0;
}
